/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    Used with ESP32 + BN220, modified by Julien Launay 14/09/2020
    Using https://github.com/khancyr/TTGO_T_BEAM/tree/master/src. Thanks to Pierre Khancyr the original author of this project.
    With https://github.com/f5soh/balise_esp32/blob/droneID_FR_testing/droneID_FR.h
    Connect ESP32 Pin16 (Rx)-> BN220 white wire (Tx)
    Connect ESP32 Pin17 (Tx)-> BN220 green wire (Rx)
	
	modif de xav:
	deux balises prises en compte: esp01 (version normale et s) et esp32
	bibliothèque droneID_FR.h commune aux deux aussi
 */

//choix de la carte entre esp32 ou esp01
#define CARTE 'esp01'
//#define CARTE 'esp32'

#if CARTE == 'esp01'
  #include <ESP8266WiFi.h>
  #include <SoftwareSerial.h>  
  extern "C" 
  {
    #include "user_interface.h"
    int wifi_send_pkt_freedom(uint8 *buf, int len, bool sys_seq);
    int  wifi_set_user_fixed_rate(uint8  enable_mask,  uint8 rate);
    //int  wifi_get_user_fixed_rate(uint8  *enable_mask, uint8 *rate);
    bool  wifi_set_user_rate_limit(uint8  mode, uint8 ifidx,  uint8 max, uint8 min);
    bool  wifi_set_user_limit_rate_mask(uint8 enable_mask);
    //uint8  wifi_get_user_limit_rate_mask(void);
    int  wifi_set_user_sup_rate(uint8  min,  uint8 max);
  }
  #define LED 2 // led bleue (pour le esp-01s c'est la pin 2 (si esp-01 c'est pin 1, conflit avec le serial, donc mettre ici genre 4 pour désactiver cette fonction)
  #define ON LOW
  #define OFF HIGH
  #define GPS_RX_PIN 0       // Brancher le fil Tx du GPS
  #define GPS_TX_PIN 2       // Rx gps, pas utilisé
  SoftwareSerial Serial2(GPS_RX_PIN,GPS_TX_PIN);
#elif CARTE == 'esp32'
  #include <WiFi.h>
  extern "C"
  {
    #include "esp_wifi.h"
    #include "esp_wifi_internal.h"
    esp_err_t esp_wifi_80211_tx(wifi_interface_t ifx, const void *buffer, int len, bool en_sys_seq);
    //esp_wifi_internal_set_fix_rate, WIFI_PHY_RATE_6M 
    esp_err_t esp_wifi_internal_set_fix_rate(wifi_interface_t ifx, bool en, wifi_phy_rate_t rate);
  }
//  #define CONFIG_ESP32_WIFI_AMPDU_TX_ENABLED 0
//  #define CONFIG_ESP32_WIFI_AMPDU_RX_ENABLED 0
  #define LED 2 
  #define ON HIGH
  #define OFF LOW
  #define GPS_RX_PIN 16
  #define GPS_TX_PIN 17
#else 
  #error "définition de carte invalide"
#endif

#define GPS_BAUD_RATE 9600

#include <TinyGPS++.h>
#include "droneID_FR.h"

TinyGPSPlus gps;
TinyGPSCustom pdop(gps, "GPGSA", 15); // $GPGSA sentence, 15th element
droneIDFR drone_idfr;


/********************************************************************************************************************
 * MODIFIEZ LES VALEURS ICI
 ********************************************************************************************************************/
// Set these to your desired credentials.
/**
  * Le nom du point d'acces wifi CHANGEZ LE par ce que vous voulez !!!
  */
  /*construire le nom du reseau wifi avec l'adresse mac recuperee */
  /* on définie d'abord le préfixe du nom: */
const char prefixe_ssid[] = "BALISE_DGAC-";
// déclaration de la variable qui va contenir le ssid complet
char ssid[32];

// Mot de pass du wifi
const char *password = "1234567890";
/**
  * CHANGEZ l'ID du drone par celui que Alphatango vous a fourni (Trigramme + Modèle + numéro série) !
  ILLEGAL_DRONE_APPELEZ_POLICE17
  */
const char drone_id[] = "ILLEGAL_DRONE_APPELEZ_POLICE17";

/********************************************************************************************************************
 * XAV
 ********************************************************************************************************************/
 // beacon frame definition
static constexpr uint16_t MAX_BEACON_SIZE = 40 + 32 + droneIDFR::FRAME_PAYLOAD_LEN_MAX;  // default beaconPacket size + max ssid size + max drone id frame size
uint8_t beaconPacket[MAX_BEACON_SIZE] = {
        0x80, 0x00,                         // 0-1: Frame Control
        0x00, 0x00,                         // 2-3: Duration
        0xff, 0xff, 0xff, 0xff, 0xff, 0xff,       // 4-9: Destination address (broadcast)
        0x24, 0x62, 0xab, 0xdd, 0xb0, 0xbd,       // 10-15: Source address FAKE  // TODO should bet set manually
        0x24, 0x62, 0xab, 0xdd, 0xb0, 0xbd,       // 16-21: Source address FAKE
        0x00, 0x00,                         // 22-23: Sequence / fragment number (done by the SDK)
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, // 24-31: Timestamp (GETS OVERWRITTEN TO 0 BY HARDWARE)
        0xB8, 0x0B,                         // 32-33: Beacon interval: set to 3s == 3000TU== BB8, bytes in reverse order  // TODO: manually set it
        0x21, 0x04,                         // 34-35: Capability info
        0x03, 0x01, 0x06,                   // 36-38: DS Parameter set, current channel 6 (= 0x06), // TODO: manually set it
        0x00, 0x20,                             // 39-40: SSID parameter set, 0x20:maxlength:content
                                                        // 41-XX: SSID (max 32)
};

// activer ou non le DEBUG
boolean DEBUG = true;       
const int ClignoteErreur = 25;
boolean IsErreur = false;
// si erreur sur gps, on pause un certain temps et on retente, pas la peine d'interroger toutes les millisecondes...
// mis à 0 car sinon problème de lecture du gps avec les délais...
const int delayErreur = 0;
double codeinfo=0;

// dès que la position home a été mise, on passe à true et on y reste
bool has_set_home = false;
double home_alt = 0;
//les positions précédentes:
double lat_prev = 0;
double lon_prev = 0;
int16_t alt_prev = 0;
int16_t dir_prev = 0;
double vit_prev = 0;

int8_t P1;
uint64_t beaconSec = 0;

// lecture des données du GPS, la led s'allume puis s'éteint brièvement quand on est dans cette fonction 
//du coup elle reste bien allumée car on y passe souvent...
static void RecupDonneesGPS()
{
  // si on est en erreur, on laisse la led éteinte
  // sinon on l'allume
	if (!IsErreur) digitalWrite(LED, ON);
  //récup des données gps
	while (Serial2.available()) { gps.encode(Serial2.read()); } 
  // on n'écrit rien ça sert à rien, même en débug
	//if (DEBUG) Serial.println("... Récupération données du GPS ... ");
}

static void EnvoiTrame(double lat_actu,double lon_actu, int16_t alt_actu, int16_t dir_actu, double vit_actu)
{
    //on définit le code info dans la trame:
    drone_idfr.set_codeinfo(codeinfo);
    //on envoie les données à la biblio pour fomattage:
    drone_idfr.set_current_position(lat_actu, lon_actu, alt_actu);
    drone_idfr.set_heading(dir_actu);
    drone_idfr.set_ground_speed(vit_actu);
    drone_idfr.set_heigth(alt_actu - home_alt);
    
     /**
     * On regarde s'il est temps d'envoyer la trame d'identification drone: soit toutes les 3s soit si le drones s'est déplacé de 30m en moins de 3s.
     *  - soit toutes les 3s,
     *  - soit si le drone s'est déplacé de 30m en moins de 30s soit 10m/s ou 36km/h,
     *  - uniquement si la position Home est déjà définie,
     *  - et dans le cas où les données GPS sont nouvelles.
     */
     
    // on envoie une trame dès qu'il est temps d'en envoyer une
    if (drone_idfr.time_to_send())
    {
      // si aucune erreur et que la trame est la bonne:
      if (codeinfo==9)
      {
        // LED allumée en fonctionnement normal,
        // deux brefs clignotements pendant l'envoi de la trame
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
      }
        //On commence par renseigner le ssid du wifi dans la trame
        // write new SSID into beacon frame
        const size_t ssid_size = (sizeof(ssid)/sizeof(*ssid)) - 1; // remove trailling null termination
        beaconPacket[40] = ssid_size;  // set size
        memcpy(&beaconPacket[41], ssid, ssid_size); // set ssid
        const uint8_t header_size = 41 + ssid_size;  //TODO: remove 41 for a marker
        
         //On génère la trame wifi avec l'identification
        const uint8_t to_send = drone_idfr.generate_beacon_frame(beaconPacket, header_size);  // override the null termination

        // Ici on ecrit sur le port USB les données GPS pour visualisation seulement.
          Serial.println("Temps UTC: " + String(gps.date.day()) + "/" + String(gps.date.month()) + "/"  + String(gps.date.year()) + " " + String(gps.time.hour()) + ":" + String(gps.time.minute()) + ":"  + String(gps.time.second()));
          Serial.println("Position de départ: " + drone_idfr.get_home_position());
          Serial.println("Position actuelle : " + drone_idfr.get_position_actuelle());
          Serial.println("Nb de satellites: " + String(gps.satellites.value()));
          Serial.println("pdop: " + String(pdop.value()));
          Serial.println("Code info: " + String(codeinfo));
          Serial.println("--------------------------");
          
         //On envoie la trame
        #if CARTE == 'esp01'
          //puissance max
          //WiFi.setOutputPower(20.0); // max 20.5dBm
          wifi_send_pkt_freedom(beaconPacket, to_send, 0);
          //puissance min
          //WiFi.setOutputPower(0.0); // max 20.5dBm
        #elif CARTE == 'esp32'
          //puissance max
          //ESP_ERROR_CHECK ( esp_wifi_set_max_tx_power(80));
          // vérifier de la puissance d'émission
          //ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&P1));
          //Serial.print("Puissance d'émission (dBm) (20 lors de l'émission de la trame): ");Serial.println(P1*0.25);
          ESP_ERROR_CHECK(esp_wifi_80211_tx(WIFI_IF_AP, beaconPacket, to_send, true));
          //puissance min
          //ESP_ERROR_CHECK ( esp_wifi_set_max_tx_power(8));
          // vérifier de la puissance d'émission
          //ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&P1));
          //Serial.print("Puissance d'émission (dBm) (doit être minimale en dehors de l'émission de la trame): ");Serial.println(P1*0.25);
        #else 
          #error "définition de carte invalide"
        #endif
        //on désactive le pouvoir
        

        //On reset la condition d'envoi
        drone_idfr.set_last_send();
        
      // si aucune erreur et que la trame est la bonne:
      if (codeinfo==9)
      {
        // LED allumée en fonctionnement normal,
        // deux brefs clignotements pendant l'envoi de la trame
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
        delay(25);
        digitalWrite(LED, ON);
        delay(25);
        digitalWrite(LED, OFF);
      }
    }
}

// Le wifi est sur le channel 6 conformement à la spécification
static constexpr uint8_t wifi_channel = 6;

// Ensure the AP SSID is max 31 letters
// 31 lettres maxi selon l'api, 17 caractères de l'adresse mac, reste 15 pour ceux de la chaine du début moins le caractère de fin de chaine ça fait 14, 14+17=31
static_assert((sizeof(prefixe_ssid)/sizeof(*prefixe_ssid))<=(14+1), "Prefix of AP SSID should be less than 14 letters");

// Ensure the drone_id is max 30 letters
static_assert((sizeof(drone_id)/sizeof(*drone_id))<=31, "Drone ID should be less that 30 letters !");  // 30 lettres + null termination

/**
 * Phase de configuration.
 */
void setup()
{
    Serial.begin(115200);
    delay(5000);
    
    //initialisation du GPS. Bien penser à croiser Rx et Tx entre ESPxx et le GPS!
    #if CARTE == 'esp01'
      Serial2.begin(GPS_BAUD_RATE);
    #elif CARTE == 'esp32'
      Serial2.begin(GPS_BAUD_RATE, SERIAL_8N1, GPS_RX_PIN, GPS_TX_PIN);
    #else 
      #error "définition de carte invalide"
    #endif
    
    //définition de la led bleue
    pinMode(LED, OUTPUT);

  //init wifi
    Serial.println("--------------------------");
    Serial.println("Mise en place du wifi:");
    //démarrage

    WiFi.mode(WIFI_AP);
//    WiFi.mode(WIFI_OFF);
    //conversion de l'adresse mac:
    String NomAPTempo = WiFi.macAddress();
    NomAPTempo.replace(":","_");
    //concat du prefixe et de l'adresse mac
    NomAPTempo = String(prefixe_ssid) + NomAPTempo;
    //transfert dans la variable globale ssid (pour pas tout changer)
    NomAPTempo.toCharArray(ssid, 32);
    // ouf! ya surement plus simple...

     //init le wifi:
      if (WiFi.softAP(ssid, password, wifi_channel, true)==false)  // ssid, pwd, channel, hidden (facultatif), max_cnx (facultatif, de 0 à 8, 4 par défaut)
      {
        Serial.print("ECHEC de mise en place du wifi, dont le nom est: ");Serial.println(ssid);
      }
      //infos sur le wifi:
      Serial.print("Nom du point d'accès: ");Serial.println(ssid);
      //attente init correcte du wifi:
      delay(500);
    
    #if CARTE == 'esp01'
    // on met la puissance au plus bas pour consommer moins
      WiFi.setOutputPower(20.0); // max 20.5dBm
      //choix du protocole wifi 802.11g
      WiFi.setPhyMode(WIFI_PHY_MODE_11G);
      softap_config current_config;
      wifi_softap_get_config(&current_config);
      current_config.beacon_interval = 1000;
      wifi_softap_set_config(&current_config);
      //définition de la vitesse d'envoi (rate)
      wifi_set_user_fixed_rate(FIXED_RATE_MASK_ALL,  PHY_RATE_6);
      wifi_set_user_rate_limit(RC_LIMIT_11G,  1,  RATE_11G_G6M, RATE_11G_G6M);
      wifi_set_user_limit_rate_mask(LIMIT_RATE_MASK_ALL);
      wifi_set_user_sup_rate(RATE_11G6M,  RATE_11G6M);
    #elif CARTE == 'esp32'
      //choix du protocole wifi 802.11g
      //WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N
      ESP_ERROR_CHECK(esp_wifi_set_protocol(WIFI_IF_AP, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G));
      //défini la largeur de bande à 20MHz
      ESP_ERROR_CHECK(esp_wifi_set_bandwidth(WIFI_IF_AP, WIFI_BW_HT20));
      //la vitesse
      //esp_wifi_internal_set_fix_rate, WIFI_PHY_RATE_6M 
      // erreur pas encore corrigée: E (10764) wifi: can't set fix rate when tx ampdu is enabled
      //ESP_ERROR_CHECK(esp_wifi_internal_set_fix_rate(WIFI_IF_AP, true, WIFI_PHY_RATE_6M));
      //beacon intervalle 
      wifi_config_t conf_current;
      esp_wifi_get_config(WIFI_IF_AP, &conf_current);
      // Change WIFI AP default beacon interval sending to 1s.
      conf_current.ap.beacon_interval = 1000;
      ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_AP, &conf_current));
      //positionner une puissance faible:
      ESP_ERROR_CHECK ( esp_wifi_set_max_tx_power(80));
      // vérifier de la puissance d'émission
      ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&P1));
      Serial.print("Puissance d'émission (dBm): ");Serial.println(P1*0.25);
      Serial.println("--------------------------");
    #else
      #error "définition de carte invalide"
    #endif
    
    // autres paramétrages, adresse ip fixée pour avoir toujours la même:
    IPAddress Ip(192, 168, 1, 1);
    IPAddress NMask(255, 255, 255, 0);
    WiFi.softAPConfig(Ip, Ip, NMask);
  
    //affichage des infos
    IPAddress myIP = WiFi.softAPIP();
    Serial.print("Adresse IP du point d'accès: ");
    Serial.println(myIP.toString());
    Serial.print("Adresse MAC du point d'accès: ");
    Serial.println(WiFi.macAddress());
    Serial.print("Nom de la balise: ");Serial.println(drone_id);

  // définition de l'id de la balise:
    drone_idfr.set_drone_id(drone_id);
    delay(2000);
    Serial.println("--------------------------");
  
} // fin setup

/**
 * Début du code principal. C'est une boucle infinie.
 */
void loop()
{
// on attend un peu que tout s'init bien (5s à partir du démarrage de la balise, en général quand on est ici c'est déjà passé les 10s)
// ne s'exec qu'une seule fois au démarrage de la balise ce délai:
	while (millis() < 10000)
	{
		digitalWrite(LED, ON);
		Serial.print(".");
		delay(100);
		digitalWrite(LED, OFF);
		delay(100);
	}

// on récupère les données du GPS
	RecupDonneesGPS();

	if (gps.charsProcessed() < 10)
	{
		if (DEBUG) Serial.println("GPS non détecté, aucune donnée GPS reçue");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 1;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
		return;
	}

	if (!gps.location.isValid())
	{
		if (DEBUG) Serial.println("Positionnement du GPS non valide (fix en cours?)");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 2;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
    return;
	}
	if (gps.location.age()> 3000)
	{
		if (DEBUG) Serial.println("Fix du GPS perdu...");
    digitalWrite(LED, OFF);
    IsErreur = true;
		if (DEBUG) Serial.println("--------------------------");
    delay(delayErreur);
    //envoyer la trame
    codeinfo= 3;
    EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
    return;
	}

// test si précision ok, sinon on remonte au début afin de ne pas encore fixer la home position
    if (gps.satellites.value() < 4)
    {
       digitalWrite(LED, OFF);
       IsErreur = true;
       if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
       if (DEBUG) Serial.println("Nb de satellites (doit être supérieur à 3): " + String(gps.satellites.value()));
       if (DEBUG) Serial.println("--------------------------");
       delay(delayErreur);
       //envoyer la trame
       codeinfo= 4;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
       return;
    }
    //if (String(pdop.value())=="") {return;}
    if ((String(pdop.value()).toDouble() == 0.0) || (String(pdop.value())=="") || (String(pdop.value()).toDouble() > 5.0))
    {
       digitalWrite(LED, OFF);
       IsErreur = true;
       if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
       if (DEBUG) Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(pdop.value()));
       if (DEBUG) Serial.println("--------------------------");
       delay(delayErreur);
       //envoyer la trame
       codeinfo= 5;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
       return;
    }
    if (gps.altitude.meters() == 0.0)
    {
        digitalWrite(LED, OFF);
        IsErreur = true;
        if (DEBUG) Serial.println("La précision du GPS n'est pas bonne");
        if (DEBUG) Serial.println("Altitude non récupérée: " + String(gps.altitude.meters()));
        if (DEBUG) Serial.println("--------------------------");
        delay(delayErreur);
        //envoyer la trame
        codeinfo= 6;
       EnvoiTrame(lat_prev,lon_prev,alt_prev,dir_prev,vit_prev);
        return;
    }

     // ici on n'est plus en erreur
    IsErreur = false;
    
    // On renseigne une première et unique fois le point de démarrage dès que la précision est satisfaisante
    //
    if (!has_set_home)
    {
        Serial.println("Stockage de la position de départ");
        Serial.println("Nb de satellites (doit être supérieur à 3): " + String(gps.satellites.value()));
        Serial.println("Précision 3D (doit être inférieure à 5.0): " + String(pdop.value()));
        home_alt = gps.altitude.meters();
        //on fixe la position home:
        drone_idfr.set_home_position(gps.location.lat(), gps.location.lng(), home_alt);
  		  has_set_home = true;
        
        //led bleue rythme rock'n'roll, tout est ok!
        int compteur = 20;
        while(compteur > 0)
        {
          digitalWrite(LED, ON);
          delay(50);
          digitalWrite(LED, OFF);
          delay(50);
          compteur--;
        }
        // on lit la position home pour controle
        Serial.println("Position de départ: " + drone_idfr.get_home_position());
        Serial.println("Stockage de la position de départ OK");
        Serial.println("--------------------------");
        //on envoie une trame
        codeinfo= 7;
        EnvoiTrame(gps.location.lat(),gps.location.lng(), home_alt, gps.course.deg(), gps.speed.mps());
        return;
    }

    // si ici, pas d'erreur donc:
    codeinfo = 9;
    //envoi de la trame ici
    EnvoiTrame(gps.location.lat(),gps.location.lng(), gps.altitude.meters(), gps.course.deg(), gps.speed.mps());
    //on stocke les dernières positions
    lat_prev = gps.location.lat();
    lon_prev = gps.location.lng();
    alt_prev = gps.altitude.meters();
    dir_prev = gps.course.deg();
    vit_prev = gps.speed.mps();
}
